@@ -13,6 +13,11 @@ from serial import Serial, SerialException
 from serial.tools.list_ports import comports
 from multiuserfilelock import MultiUserFileLock, tmpdir, Timeout
 from pathlib import Path
+import threading
+import signal
+import logging
+from functools import wraps
+from threading import RLock
 
 locks_dir = Path(tmpdir) / 'pyilluminate'
 
@@ -54,6 +59,37 @@ known_mac_addresses = {
 MAX_ARGUMENT_CHAR_COUNT = 64 * 1
 
 
+class DelayedKeyboardInterrupt:
+    def __enter__(self):
+        self.signal_received = False
+        self._is_main_thread = threading.current_thread() is \
+            threading.main_thread()
+        if self._is_main_thread:
+            # the signal api is only available to the main thread in python
+            self.old_handler = signal.signal(signal.SIGINT, self.handler)
+
+    def handler(self, sig, frame):
+        self.signal_received = (sig, frame)
+        logging.debug('SIGINT received. Delaying KeyboardInterrupt.')
+
+    def __exit__(self, type, value, traceback):
+        # During the cleanup of objects, the signal handler of the main
+        # Thread might have already been set to None or something
+        if self._is_main_thread and self.old_handler is not None:
+            signal.signal(signal.SIGINT, self.old_handler)
+        if self.signal_received:
+            self.old_handler(*self.signal_received)
+
+
+def with_thread_lock(func):
+    @wraps(func)
+    def wrapper(self, *args, **kwargs):
+        with self._thread_lock:
+            with DelayedKeyboardInterrupt():
+                return func(self, *args, **kwargs)
+    return wrapper
+
+
 def get_port_serial_number(port):
     coms = comports()
     for c in coms:
@@ -94,7 +130,8 @@ class Illuminate:
         '_has_autoupdate',
         '_autoclear',
         '_autoupdate',
-        '_scale_factor'
+        '_scale_factor',
+        '_thread_lock',
     ]
     _known_device = known_devices
     _known_serial_numbers = known_serial_numbers
@@ -217,6 +254,8 @@ class Illuminate:
         self._led_cache = []
         self._maximum_current = maximum_current
 
+        self._thread_lock = RLock()
+
         # Create default values for device parameters
         self._color = (0, 0, 0)
         self._mac_address = ""
@@ -271,6 +310,7 @@ class Illuminate:
         """The human readable name of the device."""
         return self._device_name
 
+    @with_thread_lock
     def _load_parameters(self) -> None:
         """Read the parameters from the illuminate board.
 
@@ -312,6 +352,7 @@ class Illuminate:
         led_state['serial_number'] = self.serial_number
         self._led_state = led_state
 
+    @with_thread_lock
     def _read_led_positions(self):
         # This method is defined so that other creators of LED boards
         # Can redfine it as necessary to optimize various aspects.
@@ -372,6 +413,7 @@ class Illuminate:
     def __exit__(self, *args):
         self.close()
 
+    @with_thread_lock
     def _find_port_number(self):
         if self.serial.port is None:
             if self.serial_number is None:
@@ -387,6 +429,7 @@ class Illuminate:
             self.serial.port = port
             self.serial_number = serial_number
 
+    @with_thread_lock
     def open(self) -> None:
         if not self.serial.isOpen():
             self._find_port_number()
@@ -424,6 +467,7 @@ class Illuminate:
                 self._lock_release()
                 raise e
 
+    @with_thread_lock
     def _open(self) -> None:
         """Open the serial port. Only useful if you closed it."""
         if not self.serial.isOpen():
@@ -450,6 +494,7 @@ class Illuminate:
         # get/set firmware constants
         self.maximum_current = self._maximum_current
 
+    @with_thread_lock
     def _open_startup_procedure(self):
         sleep(0.1)
 
@@ -529,6 +574,7 @@ class Illuminate:
     def __del__(self):
         self.close()
 
+    @with_thread_lock
     def close(self) -> None:
         """Force close the serial port."""
         if self.serial is not None and self.serial.isOpen():
@@ -583,6 +629,7 @@ class Illuminate:
             self._lock.release()
             self._lock = None
 
+    @with_thread_lock
     def write(self, data) -> None:
         """Write data to the port.
 
@@ -594,6 +641,7 @@ class Illuminate:
             data = (data + '\n').encode('utf-8')
         self.serial.write(data)
 
+    @with_thread_lock
     def read(self, size: int=10000) -> bytearray:
         """Read data from the serial port.
 
@@ -607,6 +655,7 @@ class Illuminate:
             raise RuntimeError("__init__ must be successfully called first")
         return self.serial.read(size)
 
+    @with_thread_lock
     def readline(self) -> str:
         """Call underlying readline and decode as utf-8."""
         if self.serial is None:
@@ -614,6 +663,7 @@ class Illuminate:
         b = self.serial.readline()
         return b.decode('utf-8')
 
+    @with_thread_lock
     def read_paragraph(self, raw=False) -> List[str]:
         """Read a whole paragraph of text.
 
@@ -643,6 +693,7 @@ class Illuminate:
             if '-==-' in line_stripped:
                 return paragraph
 
+    @with_thread_lock
     def ask(self, data: str) -> Union[int, float, None]:
         """Send data, read the output, check for error, extract a number."""
         p = self._ask_list(data)
@@ -686,11 +737,13 @@ class Illuminate:
         else:
             return None
 
+    @with_thread_lock
     def _ask_list(self, data: str, raw: bool=False) -> List[str]:
         """Read data, return as list of strings."""
         self.write(data)
         return self.read_paragraph(raw)
 
+    @with_thread_lock
     def _ask_string(self, data: str, raw: bool =False) -> str:
         """Read data, return as a single string."""
         p = self._ask_list(data, raw)
@@ -754,6 +807,7 @@ class Illuminate:
         return self._autoclear
 
     @autoclear.setter
+    @with_thread_lock
     def autoclear(self, value: bool=None) -> None:
         # The autoclear command from the teensy toggles the
         # autoclear bit, so we must remember the state of autoclear
@@ -780,6 +834,7 @@ class Illuminate:
         return self._autoupdate
 
     @autoupdate.setter
+    @with_thread_lock
     def autoupdate(self, value: bool=None) -> None:
         # The autoupdate command from the teensy toggles the
         # autoupdate bit, so we must remember the state of autoupdate
@@ -840,6 +895,7 @@ class Illuminate:
         return user_color
 
     @color.setter
+    @with_thread_lock
     def color(self, c: Union[float, Iterable[float]]):
         if not isinstance(c, collections.abc.Iterable):
             # Make it a 3 tuple
@@ -944,6 +1000,7 @@ class Illuminate:
         raise AttributeError(
             "The ``leds`` attribute doesn't exist. Did you mean ``led``")
 
+    @with_thread_lock
     def turn_on_led(self, leds: Union[int, Iterable[int]]) -> None:
         """Turn on a single LED(or multiple LEDs in an iterable).
 
@@ -1018,6 +1075,7 @@ class Illuminate:
     def _led(self, value):
         self._update_led_state(value)
 
+    @with_thread_lock
     def _update_led_state(self, led, force_clear=False):
         # The LED state might not exist if the system hasn't booted up
         # Therefore, skip any errors that may occur on "Unefined variable"
@@ -1041,6 +1099,7 @@ class Illuminate:
         self._led_state.data[led] = color
         self._led_cache = led_cache
 
+    @with_thread_lock
     def clear(self) -> None:
         """Clear the LED array."""
         self.ask('x')
@@ -1054,6 +1113,7 @@ class Illuminate:
                 "Contact support for help.")
         self.ask('u')
 
+    @with_thread_lock
     def fill_array(self) -> None:
         """Fill the LED array with default color."""
         self.ask('ff')
@@ -1411,6 +1471,7 @@ class Illuminate:
         j = json.loads(self._ask_string('pledposna'))
         return j['led_position_list_na']
 
+    @with_thread_lock
     def discoparty_demo(self, n_leds=1, time=10):
         """Run a demo routine to show what the array can do.
 
@@ -1426,6 +1487,7 @@ class Illuminate:
         self.write('disco.' + str(n_leds))
         self._finish_demo(time)
 
+    @with_thread_lock
     def demo(self, time: float=10) -> None:
         """Run a demo routine to show what the array can do.
 
@@ -1449,6 +1511,7 @@ class Illuminate:
             raise inst
         self.serial.timeout = previous_timeout
 
+    @with_thread_lock
     def water_drop_demo(self, time: float=10) -> None:
         """Water drop demo."""
         # Basically, if you let this one go on, and it actually returns
@@ -1457,6 +1520,7 @@ class Illuminate:
         self.write('water')
         self._finish_demo(time)
 
+    @with_thread_lock
     def _finish_demo(self, time: float) -> None:
         if self.serial is None:
             raise RuntimeError("__init__ must be successfully called first")
